"""
Store metadata in BibTeX.
"""

import json
from typing import List, Union

import bibtexparser
from bibtexparser import Library
from bibtexparser.model import Entry, Field
from langcodes import Language


def normalize(text: str) -> str:
    return text.replace("{", "").replace("}", "")


def get_author(d) -> Union[str, None]:
    if d["authors"] is None:
        return None
    return " and ".join(d["authors"])


def get_year(d) -> Union[str, None]:
    if isinstance(d["publishDate"], list):
        return f'{d["publishDate"][0]["year"]}--{d["publishDate"][1]["year"]}'
    if d["publishDate"] is None:
        return None
    return f'{d["publishDate"]["year"]}'


def get_keywords(d) -> Union[str, None]:
    if len(d["tags"]) == 0:
        return None
    return ", ".join(d["tags"])


def get_languages(d) -> Union[str, None]:
    if d["languages"] is None:
        return None
    languages = [Language.get(lang).language_name() for lang in d["languages"]]
    return ", ".join(languages)


def build_fields(d) -> List[Field]:
    fields = [
        Field("url", d["viewUrl"]),
        Field("file", d["downloadUrl"]),
    ]

    if d["displayName"] is not None:
        fields.append(Field("title", d["displayName"]))

    author = get_author(d)
    if author is not None:
        fields.append(Field("author", author))

    year = get_year(d)
    if year is not None:
        fields.append(Field("year", year))

    keywords = get_keywords(d)
    if keywords is not None:
        fields.append(Field("keywords", keywords))

    languages = get_languages(d)
    if languages is not None:
        fields.append(Field("languages", languages))

    if d["abstract"] is not None:
        fields.append(Field("abstract", d["abstract"]))

    for field in fields:
        field.value = normalize(field.value)
    return fields


def save_as_bib(metadata_path: str, bib_path: str) -> None:
    with open(metadata_path, "r", encoding="utf-8") as f:
        visualizations = json.load(f)
    library = Library(
        [
            Entry(entry_type="misc", key=d["uuid"], fields=build_fields(d))
            for d in visualizations
        ]
    )
    with open(bib_path, "w", encoding="utf-8") as f:
        f.write(bibtexparser.write_string(library))



import json
from typing import Any

import requests
from requests.exceptions import ConnectionError
from tqdm import tqdm


def download(url: str, path: str) -> None:
    """Download file from URL to path."""

    # The received content length would be
    # the size of compressed file without Accept-Encoding: identity.
    r = requests.head(url, headers={"Accept-Encoding": "identity"}, timeout=20)
    content_length = int(r.headers["Content-Length"])

    r = requests.get(url, stream=True, timeout=20)
    with open(path, "wb") as f, tqdm(unit="B", total=content_length) as pbar:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                pbar.update(len(chunk))
                f.write(chunk)


class Dataset:
    """
    The dataset class for downloading and loading remote datasets.
    """

    def __init__(self, name: str, url: str, format: str):
        self.name = name
        self.url = url
        self.format = format

    def load(self) -> Any:
        if self.format == "json":
            return json.loads(requests.get(self.url).content)
        raise ValueError(
            f"Unrecognized file format: {self.format}. Valid options are ['json']."
        )

    def download(self, path: str) -> None:
        try:
            download(self.url, path)
        except (ConnectionError, TimeoutError) as e:
            print("Download failed because of network connection issue.")
            raise e



"""
Utility functions to download images from the urls stored in metadata.
"""

import json
import mimetypes
import os
from os import listdir
from os.path import isfile, join
from typing import List, TypedDict

import requests
from tqdm import tqdm


class ImageQuery(TypedDict):
    """
    The data structure of an image query.

    Attributes
    ----------
    url : str
        The url for fetching the image.
    uuid : str
        The UUID of the image.
    """

    url: str
    uuid: str


def filename2uuid(filename: str) -> str:
    return filename.split(".")[0]


def filter_queries(img_queries: List[ImageQuery], img_dir: str) -> List[ImageQuery]:
    """
    Filter URLs queried before according to the stored images.
    """

    filenames = [d for d in listdir(img_dir) if isfile(join(img_dir, d))]
    img_uuids = {filename2uuid(d) for d in filenames}
    return [d for d in img_queries if d["uuid"] not in img_uuids]


def fetch_images(
    metadata_path: str,
    img_dir: str,
) -> None:
    """
    Given base urls, generate image queries, and store the query results.

    Args
    ----
    metadata_path : str
        The path to the metadata file where image urls are stored.
    img_dir : str
        The path to the directory for storing the images.
    """

    if not os.path.exists(img_dir):
        os.makedirs(img_dir)

    s = requests.Session()

    with open(metadata_path, "r", encoding="utf-8") as f:
        metadata = json.load(f)

    queries = filter_queries(metadata, img_dir)
    for query in tqdm(queries, desc="Fetch Image Progress"):
        uuid = query["uuid"]
        response = s.get(query["downloadUrl"])
        extension = mimetypes.guess_extension(response.headers["content-type"])
        with open(f"{img_dir}/{uuid}{extension}", "wb") as f:
            f.write(response.content)



from .bib import save_as_bib
from .images import fetch_images
from .metadata import Dataset

visualizations = Dataset(
    name="visualizations",
    url="https://raw.githubusercontent.com/oldvis/dataset/main/dataset/output/visualizations.json",
    format="json",
)
authors = Dataset(
    name="authors",
    url="https://raw.githubusercontent.com/oldvis/dataset/main/dataset/output/authors.json",
    format="json",
)

__all__ = [
    "save_as_bib",
    "Dataset",
    "visualizations",
    "authors",
    "fetch_images",
]
